#![allow(dead_code)]

use std::any::{Any, TypeId};

use component::Component;
use component_macro::Component;

mod component;

struct Context<'a> {
    entities: &'a mut Vec<(u64, Vec<Box<dyn Component>>)>,
}

impl<'a> Context<'a> {
    pub fn entities_with_component<T: 'static>(&self) -> Vec<u64> {
        let entity_type_id = TypeId::of::<T>();
        self.entities
            .iter()
            .filter_map(|(id, components)| {
                let contains_component = components.iter().any(|entity| {
                    let is_id = (*entity).inner_type_id() == entity_type_id;
                    is_id
                });
                if contains_component {
                    Some(*id)
                } else {
                    None
                }
            })
            .collect()
    }

    pub fn entity_component<T: 'static>(&mut self, entity_id: u64) -> &mut T {
        let entity_type_id = TypeId::of::<T>();
        let (_id, components) = self
            .entities
            .iter_mut()
            .find(|(id, _)| *id == entity_id)
            .unwrap();

        let component = components
            .iter_mut()
            .find_map(|entity| {
                let is_id = (*entity).inner_type_id() == entity_type_id;
                if is_id {
                    Some(entity.as_any().downcast_mut::<T>().unwrap())
                } else {
                    None
                }
            })
            .unwrap();
        component
    }
}

trait System {
    fn on_update(&self, context: &mut Context);
}

#[derive(Component)]
struct Position {
    pub x: i32,
    pub y: i32,
}

#[derive(Component)]
struct NotPosition {
    pub a: i32,
    pub b: i32,
}

struct Gravity;
struct NotGravity;

impl System for Gravity {
    fn on_update(&self, context: &mut Context) {
        let entities_with_position = context.entities_with_component::<Position>();
        for entity_id in entities_with_position {
            let position = context.entity_component::<Position>(entity_id);
            position.y -= 1;
        }
    }
}

impl System for NotGravity {
    fn on_update(&self, context: &mut Context) {
        let entities_with_position = context.entities_with_component::<NotPosition>();
        for entity_id in entities_with_position {
            let position = context.entity_component::<NotPosition>(entity_id);
            position.a -= 1;
        }
    }
}

fn main() {
    let mut entities: Vec<(u64, Vec<Box<dyn Component>>)> = Vec::new();
    let mut systems: Vec<Box<dyn System>> = Vec::new();

    systems.push(Box::new(Gravity));
    systems.push(Box::new(NotGravity));
    entities.push((
        0,
        vec![
            Box::new(Position { x: 0, y: 0 }),
            Box::new(NotPosition { a: 0, b: 0 }),
        ],
    ));

    entities.push((
        1,
        vec![
            Box::new(Position { x: 0, y: 0 }),
            Box::new(NotPosition { a: 0, b: 0 }),
        ],
    ));

    let mut context = Context {
        entities: &mut entities,
    };

    for system in systems {
        system.on_update(&mut context)
    }
}
